#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

unsigned long utime(void); // for benchmarking

// result of computation is a square grid of integers with this
// many rows and columns of elements
#define DIM 800

// Number of threads per block: should be a multiple of 32.
//#define THREADS_PER_BLOCK 32
#define THREADS_PER_BLOCK 1

struct cuComplex {
	float r;
	float i;
};

__device__ cuComplex cuComplexMake(double r, double i)
{
	cuComplex c;
	c.r = r;
	c.i = i;
	return c;
}

__device__ float cuComplexMagnitude( cuComplex c )
{
	return sqrt( c.r * c.r + c.i * c.i );
}

__device__ cuComplex cuComplexMultiply(cuComplex lhs, cuComplex rhs)
{
	return cuComplexMake(
		lhs.r * rhs.r - lhs.i * rhs.i,
		lhs.i * rhs.r + lhs.r * rhs.i
	);
}

__device__ cuComplex cuComplexAdd(cuComplex lhs, cuComplex rhs)
{
	return cuComplexMake( lhs.r + rhs.r, lhs.i + rhs.i );
}

__device__ int cuMandelIters( cuComplex c )
{
	cuComplex z;
	int iters = 0;

	z = cuComplexMake( 0.0f, 0.0f );
	while ( iters < 1000 ) {
		z = cuComplexMultiply( z, z );
		z = cuComplexAdd( z, c );
		if ( cuComplexMagnitude( z ) > 2.0f ) {
			break;
		}
		iters++;
	}

	return iters;
}

__global__ void kernel( int *data )
{
	int i = blockIdx.x;
	int j = blockIdx.y*blockDim.x + threadIdx.x;

	// If the number of rows (DIM) did not divide evenly by the
	// number of threads, then the last "row" of blocks will
	// have some excess threads that won't be assigned to
	// actual rows of the computation.  The if statement checks
	// to make sure that the grid row index (j) is valid
	// before doing the computation.

	if (j < DIM) {
		float x = -2.0f + (((float)i / DIM) * 4.0f);
		float y = -2.0f + (((float)j / DIM) * 4.0f);

		int iters = cuMandelIters( cuComplexMake( x, y ) );

		data[j * DIM + i ] = iters;
	}
}

int main(void)
{
	int *host_data, *dev_data;

	unsigned long start = utime();

	const size_t bufSize = (DIM*DIM) * sizeof(int);

	host_data = (int *) malloc( bufSize );
	cudaMalloc( (void**) &dev_data, bufSize );

	int num_blocks_y = DIM / THREADS_PER_BLOCK; // determine how many rows per block
	if (DIM % THREADS_PER_BLOCK > 0) {
		// Number of rows did not divide evenly by number of threads:
		// add one additional block.  The kernel function will
		// check to make sure that the block index and the thread
		// index yield a valid row.
		num_blocks_y++;
	}
	dim3 grid(DIM, num_blocks_y);
	kernel<<<grid, THREADS_PER_BLOCK>>>( dev_data );

	cudaMemcpy( host_data, dev_data, bufSize, cudaMemcpyDeviceToHost );

	unsigned long end = utime();

	printf("Computation finished in %lu us\n", (end-start));

	FILE *out = fopen("result.txt", "w");

	for (int j = 0; j < DIM; j++) {
		for (int i = 0; i < DIM; i++) {
			if (i > 0) {
				fprintf(out, " ");
			}
			fprintf(out, "%d", host_data[j*DIM + i]);
		}
		fprintf(out, "\n");
	}
	fclose(out);

	printf("Success?\n");

	return 0;
}

unsigned long utime(void)
{
	struct timeval tv;
	unsigned long result = 0;

	gettimeofday(&tv, NULL);
	result += (tv.tv_sec * 1000000);
	result += tv.tv_usec;

	return result;
}

// vim: syntax=c
